import logging
from typing import Union

from spaceone.core.service import *
from spaceone.core.service.utils import *
from spaceone.core.error import *

from spaceone.identity.manager.provider_manager import ProviderManager
from spaceone.identity.manager.schema_manager import SchemaManager
from spaceone.identity.manager.trusted_account_manager import TrustedAccountManager
from spaceone.identity.manager.workspace_manager import WorkspaceManager
from spaceone.identity.manager.secret_manager import SecretManager
from spaceone.identity.model.trusted_account.request import *
from spaceone.identity.model.trusted_account.response import *
from spaceone.identity.model.provider.database import Provider
from spaceone.identity.model.job.response import JobResponse
from spaceone.identity.service.job_service import JobService

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class TrustedAccountService(BaseService):
    resource = "TrustedAccount"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.trusted_account_mgr = TrustedAccountManager()
        self.provider_mgr = ProviderManager()

    @transaction(
        permission="identity:TrustedAccount.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER"],
    )
    @convert_model
    def create(
        self, params: TrustedAccountCreateRequest
    ) -> Union[TrustedAccountResponse, dict]:
        """create trusted account

         Args:
            params (TrustedAccountCreateRequest): {
                'name': 'str',                  # required
                'data': 'dict',                 # required
                'provider': 'str',              # required
                'secret_schema_id': 'str',
                'secret_data': 'dict',
                'schedule': 'dict',
                'sync_options': 'dict',
                'plugin_options': 'dict',
                'tags': 'dict',
                'resource_group': 'str',        # required
                'workspace_id': 'str',          # injected from auth
                'domain_id': 'str'              # injected from auth (required)
            }

        Returns:
            TrustedAccountResponse:
        """

        # Check workspace
        if params.resource_group == "WORKSPACE":
            if params.workspace_id is None:
                raise ERROR_REQUIRED_PARAMETER(key="workspace_id")

            workspace_mgr = WorkspaceManager()
            workspace_mgr.get_workspace(params.workspace_id, params.domain_id)
        else:
            params.workspace_id = "*"

        # Check provider
        if params.schedule or params.sync_options or params.plugin_options:
            provider_vo = self.provider_mgr.get_provider(
                params.provider, params.domain_id
            )
            self._check_provider_sync(provider_vo)

        # Check data by schema
        schema_mgr = SchemaManager()
        schema_mgr.validate_data_by_schema(
            params.provider, params.domain_id, "TRUSTED_ACCOUNT", params.data
        )

        trusted_account_vo = self.trusted_account_mgr.create_trusted_account(
            params.dict()
        )

        # Check secret_data by schema
        schema_mgr.validate_secret_data_by_schema_id(
            params.secret_schema_id, params.domain_id, params.secret_data, "SECRET"
        )

        # Create a trusted secret
        secret_mgr = SecretManager()
        trusted_secret_info = secret_mgr.create_trusted_secret(
            {
                "name": f"{trusted_account_vo.trusted_account_id}-trusted-secret",
                "data": params.secret_data,
                "schema_id": params.secret_schema_id,
                "trusted_account_id": trusted_account_vo.trusted_account_id,
                "resource_group": params.resource_group,
                "workspace_id": params.workspace_id,
            }
        )

        # Update trusted_secret_id in trusted_account_vo
        trusted_account_vo = self.trusted_account_mgr.update_trusted_account_by_vo(
            {"trusted_secret_id": trusted_secret_info["trusted_secret_id"]},
            trusted_account_vo,
        )

        return TrustedAccountResponse(**trusted_account_vo.to_dict())

    @transaction(
        permission="identity:TrustedAccount.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER"],
    )
    @convert_model
    def update(
        self, params: TrustedAccountUpdateRequest
    ) -> Union[TrustedAccountResponse, dict]:
        """update trusted account

         Args:
            params (TrustedAccountUpdateRequest): {
                'trusted_account_id': 'str',    # required
                'name': 'str',
                'data': 'dict',
                'schedule': 'dict',
                'sync_options': 'dict',
                'plugin_options': 'dict',
                'tags': 'dict',
                'workspace_id': 'str',          # injected from auth
                'domain_id': 'str'              # injected from auth (required)
            }

        Returns:
            TrustedAccountResponse:
        """

        trusted_account_vo = self.trusted_account_mgr.get_trusted_account(
            params.trusted_account_id, params.domain_id, params.workspace_id
        )

        if params.data:
            # Check data by schema
            schema_mgr = SchemaManager()
            schema_mgr.validate_data_by_schema(
                trusted_account_vo.provider,
                params.domain_id,
                "TRUSTED_ACCOUNT",
                params.data,
            )

        if params.sync_options or params.schedule or params.plugin_options:
            provider_vo = self.provider_mgr.get_provider(
                trusted_account_vo.provider, params.domain_id
            )
            self._check_provider_sync(provider_vo)

        trusted_account_vo = self.trusted_account_mgr.update_trusted_account_by_vo(
            params.dict(exclude_unset=True), trusted_account_vo
        )

        return TrustedAccountResponse(**trusted_account_vo.to_dict())

    @transaction(
        permission="identity:TrustedAccount.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER"],
    )
    @convert_model
    def update_secret_data(
        self, params: TrustedAccountUpdateSecretRequest
    ) -> Union[TrustedAccountResponse, dict]:
        """update trusted account secret data

         Args:
            params (TrustedAccountUpdateSecretRequest): {
                'trusted_account_id': 'str',    # required
                'secret_schema_id': 'str',      # required
                'secret_data': 'dict',          # required
                'workspace_id': 'str',          # injected from auth
                'domain_id': 'str'              # injected from auth (required)
            }

        Returns:
            TrustedAccountResponse:
        """

        trusted_account_vo = self.trusted_account_mgr.get_trusted_account(
            params.trusted_account_id, params.domain_id, params.workspace_id
        )

        # Check secret_data by schema
        schema_mgr = SchemaManager()
        schema_mgr.validate_secret_data_by_schema_id(
            params.secret_schema_id, params.domain_id, params.secret_data, "SECRET"
        )

        # Update secret data
        secret_mgr = SecretManager()
        secret_mgr.update_trusted_secret_data(
            trusted_account_vo.trusted_secret_id,
            params.secret_schema_id,
            params.secret_data,
        )

        if trusted_account_vo.secret_schema_id != params.secret_schema_id:
            trusted_account_vo = self.trusted_account_mgr.update_trusted_account_by_vo(
                {"secret_schema_id": params.secret_schema_id}, trusted_account_vo
            )

        return TrustedAccountResponse(**trusted_account_vo.to_dict())

    @transaction(
        permission="identity:TrustedAccount.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER"],
    )
    @convert_model
    def delete(self, params: TrustedAccountDeleteRequest) -> None:
        """delete trusted account

         Args:
            params (TrustedAccountDeleteRequest): {
                'trusted_account_id': 'str',    # required
                'workspace_id': 'str',          # injected from auth
                'domain_id': 'str'              # injected from auth (required)
            }

        Returns:
            None
        """

        trusted_account_vo = self.trusted_account_mgr.get_trusted_account(
            params.trusted_account_id, params.domain_id, params.workspace_id
        )

        # Delete trusted secret
        secret_mgr = SecretManager()
        secret_mgr.delete_related_trusted_secrets(trusted_account_vo.trusted_account_id)

        self.trusted_account_mgr.delete_trusted_account_by_vo(trusted_account_vo)

    @transaction(
        permission="identity:TrustedAccount.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER"],
    )
    @convert_model
    def sync(self, params: TrustedAccountSyncRequest) -> Union[JobResponse, dict]:
        """Sync trusted account
        Args:
           params (TrustedAccountSyncRequest): {
               'trusted_account_id': 'str',    # required
               'workspace_id': 'str',          # injected from auth
               'domain_id': 'str'              # injected from auth (required)
           }
        Returns:
            JobResponse: 'dict'
        """

        job_service = JobService()

        trusted_account_id = params.trusted_account_id
        workspace_id = params.workspace_id
        domain_id = params.domain_id

        trusted_account_vo = self.trusted_account_mgr.get_trusted_account(
            trusted_account_id, domain_id, workspace_id
        )

        provider_vo = self.provider_mgr.get_provider(
            trusted_account_vo.provider, domain_id
        )
        self._check_provider_sync(provider_vo)
        job_vo = job_service.create_service_account_job(trusted_account_vo, {})
        return JobResponse(**job_vo.to_dict())

    @transaction(
        permission="identity:TrustedAccount.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @change_value_by_rule("APPEND", "workspace_id", "*")
    @convert_model
    def get(
        self, params: TrustedAccountGetRequest
    ) -> Union[TrustedAccountResponse, dict]:
        """get trusted account

         Args:
            params (TrustedAccountGetRequest): {
                'trusted_account_id': 'str',    # required
                'workspace_id': 'list',         # injected from auth
                'domain_id': 'str'              # injected from auth (required)
            }

        Returns:
            TrustedAccountResponse:
        """

        trusted_account_vo = self.trusted_account_mgr.get_trusted_account(
            params.trusted_account_id, params.domain_id, params.workspace_id
        )

        return TrustedAccountResponse(**trusted_account_vo.to_dict())

    @transaction(
        permission="identity:TrustedAccount.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @change_value_by_rule("APPEND", "workspace_id", "*")
    @append_query_filter(
        [
            "trusted_account_id",
            "name",
            "provider",
            "secret_schema_id",
            "trusted_secret_id",
            "workspace_id",
            "domain_id",
        ]
    )
    @append_keyword_filter(["trusted_account_id", "name"])
    @convert_model
    def list(
        self, params: TrustedAccountSearchQueryRequest
    ) -> Union[TrustedAccountsResponse, dict]:
        """list trusted accounts

        Args:
            params (TrustedAccountSearchQueryRequest): {
                'query': 'dict (spaceone.api.core.v1.Query)',
                'trusted_account_id': 'str',
                'name': 'str',
                'provider': 'str',
                'secret_schema_id': 'str',
                'trusted_secret_id': 'str',
                'workspace_id': 'list',     # injected from auth
                'domain_id': 'str',         # injected from auth (required)
            }

        Returns:
            TrustedAccountsResponse:
        """

        query = params.query or {}
        (
            trusted_account_vos,
            total_count,
        ) = self.trusted_account_mgr.list_trusted_accounts(query)

        trusted_accounts_info = [
            trusted_account_vo.to_dict() for trusted_account_vo in trusted_account_vos
        ]
        return TrustedAccountsResponse(
            results=trusted_accounts_info, total_count=total_count
        )

    @transaction(
        permission="identity:TrustedAccount.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @change_value_by_rule("APPEND", "workspace_id", "*")
    @append_query_filter(["workspace_id", "domain_id"])
    @append_keyword_filter(["trusted_account_id", "name"])
    @convert_model
    def stat(self, params: TrustedAccountStatQueryRequest) -> dict:
        """stat trusted accounts

        Args:
            params (TrustedAccountStatQueryRequest): {
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)', # required
                'workspace_id': 'list',     # injected from auth
                'domain_id': 'str',         # injected from auth (required)
            }

        Returns:
            dict: {
                'results': 'list',
                'total_count': 'int'
            }

        """

        query = params.query or {}
        return self.trusted_account_mgr.stat_trusted_accounts(query)

    @staticmethod
    def _check_provider_sync(provider_vo: Provider) -> None:
        options = provider_vo.options or {}
        if not options.get("support_trusted_account") and options.get(
            "support_auto_sync"
        ):
            raise ERROR_INVALID_PARAMETER(
                key="provider.options", message="Sync options is disabled"
            )
        elif not provider_vo.plugin_info:
            raise ERROR_INVALID_PARAMETER(
                key="provider.plugin_info", message="Plugin info not found"
            )



import logging

from spaceone.core.connector import BaseConnector

from spaceone.identity.error.error_authentication import *

_LOGGER = logging.getLogger(__name__)


class ExternalAuthPluginConnector(BaseConnector):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.client = None

    def initialize(self, endpoint):
        static_endpoint = self.config.get("endpoint")

        if static_endpoint:
            endpoint = static_endpoint

        _LOGGER.info(f"[initialize] endpoint: {endpoint}")
        self.client = self.locator.get_connector("SpaceConnector", endpoint=endpoint)

    def init(self, options: dict, domain_id: str):
        params = {"options": options, "domain_id": domain_id}

        try:
            return self.client.dispatch("ExternalAuth.init", params)

        except ERROR_BASE as e:
            raise ERROR_AUTHENTICATION_FAILURE_PLUGIN(message=e.message)
        except Exception as e:
            raise ERROR_AUTHENTICATION_FAILURE_PLUGIN(messsage=str(e))

    def authorize(
        self,
        credentials,
        options,
        secret_data,
        domain_id,
        schema_id=None,
        metadata=None,
    ):
        params = {
            "options": options,
            "secret_data": secret_data,
            "credentials": credentials,
            "schema_id": schema_id,
            "domain_id": domain_id,
            "metadata": metadata or {},
        }

        try:
            user_info = self.client.dispatch("ExternalAuth.authorize", params)
            return user_info
        except ERROR_BASE as e:
            _LOGGER.error(
                f"[authorize] ExternalAuth.authorize failed. (reason={e.message})"
            )
            raise ERROR_INVALID_CREDENTIALS()
        except Exception as e:
            _LOGGER.error(
                f"[authorize] ExternalAuth.authorize failed. (reason={str(e)})"
            )
            raise ERROR_INVALID_CREDENTIALS()

    # def call_find(self, keyword, user_id, options, secret_data={}, schema=None):
    #     params = {
    #         "options": options,
    #         "secret_data": secret_data,
    #         "schema": schema,
    #         "keyword": keyword,
    #         "user_id": user_id,
    #     }
    #     _LOGGER.info(f"[call_find] params: {params}")
    #
    #     try:
    #         response = self.client.Auth.find(
    #             params, metadata=self.transaction.get_connection_meta()
    #         )
    #
    #         _LOGGER.debug(f"[call_find] MessageToDict(user_info): {users_info}")
    #         return users_info
    #
    #     except ERROR_BASE as e:
    #         raise ERROR_AUTHENTICATION_FAILURE_PLUGIN(message=e.message)
    #     except Exception as e:
    #         raise ERROR_AUTHENTICATION_FAILURE_PLUGIN(messsage=str(e))



import logging
from typing import Dict, List, Tuple

from spaceone.core.error import ERROR_INVALID_PARAMETER, ERROR_PERMISSION_DENIED
from spaceone.core.manager import BaseManager

from spaceone.identity.manager.role_binding_manager import RoleBindingManager
from spaceone.identity.manager.user_manager import UserManager

_LOGGER = logging.getLogger(__name__)


class WorkspaceGroupUserManager(BaseManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.rb_mgr = RoleBindingManager()
        self.user_mgr = UserManager()

    def stat_workspace_group_users(
        self, query: dict, workspace_group_id: str, domain_id: str
    ) -> dict:
        user_ids, _ = self._get_role_binding_map_in_workspace_group(
            workspace_group_id, domain_id
        )
        query["filter"] = query.get("filter", [])
        query["filter"].append({"k": "user_id", "v": user_ids, "o": "in"})

        return self.user_mgr.stat_users(query)

    @staticmethod
    def check_user_role_type(
        old_users_in_workspace_group: List[Dict[str, str]],
        user_id: str,
    ) -> None:
        user_role_type = ""
        for old_user in old_users_in_workspace_group:
            if old_user["user_id"] == user_id:
                user_role_type = old_user["role_type"]

        if user_role_type == "WORKSPACE_MEMBER":
            _LOGGER.error(
                f"[check_user_role_type] User ID {user_id} is WORKSPACE_MEMBER."
            )
            raise ERROR_PERMISSION_DENIED()

    def _get_role_binding_map_in_workspace_group(
        self,
        workspace_group_id: str,
        domain_id: str,
        role_type: str = None,
    ) -> Tuple[list, dict]:
        user_rb_map = {}
        user_ids = []

        if role_type and role_type not in ["WORKSPACE_OWNER", "WORKSPACE_MEMBER"]:
            raise ERROR_INVALID_PARAMETER(
                message="role_type must be one of [WORKSPACE_OWNER, WORKSPACE_MEMBER]"
            )

        if role_type is None:
            role_type = ["WORKSPACE_OWNER", "WORKSPACE_MEMBER"]

        rb_vos = self.rb_mgr.filter_role_bindings(
            role_type=role_type,
            workspace_group_id=[workspace_group_id, "*"],
            domain_id=domain_id,
        )

        for rb_vo in rb_vos:
            user_ids.append(rb_vo.user_id)
            if user_rb_map.get(rb_vo.user_id):
                user_rb_map[rb_vo.user_id].append(rb_vo.to_dict())
            else:
                user_rb_map[rb_vo.user_id] = [rb_vo.to_dict()]

        return user_ids, user_rb_map



from typing import List

from pydantic.main import BaseModel

__all__ = [
    "WorkspaceGroupUserSummaryResponse",
    "WorkspaceGroupUsersSummaryResponse",
]

from spaceone.identity.model.workspace_group.request import State


class WorkspaceGroupUserSummaryResponse(BaseModel):
    user_id: str
    name: str
    state: State


class WorkspaceGroupUsersSummaryResponse(BaseModel):
    results: List[WorkspaceGroupUserSummaryResponse]
    total_count: int



from mongoengine import *
from spaceone.core.model.mongo_model import MongoModel


class Provider(MongoModel):
    provider = StringField(max_length=40, unique_with="domain_id")
    name = StringField(max_length=40)
    plugin_info = DictField(default=None, null=True)
    version = StringField(max_length=40, default=None, null=True)
    alias = StringField(max_length=40, default=None, null=True)
    color = StringField(max_length=7, default=None, null=True)
    icon = StringField(default=None, null=True)
    order = IntField(min_value=1, default=10)
    options = DictField(default=None)
    tags = DictField(default=None)
    is_managed = BooleanField(default=False)
    domain_id = StringField(max_length=40)
    created_at = DateTimeField(auto_now_add=True)
    updated_at = DateTimeField(auto_now=True)

    meta = {
        "updatable_fields": [
            "name",
            "plugin_info",
            "version",
            "alias",
            "color",
            "icon",
            "order",
            "options",
            "tags",
            "updated_at",
        ],
        "minimal_fields": ["provider", "name", "order", "is_managed"],
        "ordering": ["order", "name"],
        "indexes": ["provider", "is_managed", "domain_id"],
    }



from google.protobuf.json_format import ParseDict
from spaceone.api.identity.v2 import (user_pb2, user_profile_pb2,
                                      user_profile_pb2_grpc)
from spaceone.core.pygrpc import BaseAPI

from spaceone.identity.service.user_profile_service import UserProfileService


class UserProfile(BaseAPI, user_profile_pb2_grpc.UserProfileServicer):
    pb2 = user_profile_pb2
    pb2_grpc = user_profile_pb2_grpc

    def update(self, request, context):
        params, metadata = self.parse_request(request, context)
        user_profile_svc = UserProfileService(metadata)
        response: dict = user_profile_svc.update(params)
        return ParseDict(response, user_pb2.UserInfo())

    def verify_email(self, request, context):
        params, metadata = self.parse_request(request, context)
        user_profile_svc = UserProfileService(metadata)
        user_profile_svc.verify_email(params)
        return self.empty()

    def confirm_email(self, request, context):
        params, metadata = self.parse_request(request, context)
        user_profile_svc = UserProfileService(metadata)
        response: dict = user_profile_svc.confirm_email(params)
        return ParseDict(response, user_pb2.UserInfo())

    def reset_password(self, request, context):
        params, metadata = self.parse_request(request, context)
        user_profile_svc = UserProfileService(metadata)
        user_profile_svc.reset_password(params)
        return self.empty()

    def enable_mfa(self, request, context):
        params, metadata = self.parse_request(request, context)
        user_profile_svc = UserProfileService(metadata)
        response: dict = user_profile_svc.enable_mfa(params)
        return ParseDict(response, user_pb2.UserInfo())

    def disable_mfa(self, request, context):
        params, metadata = self.parse_request(request, context)
        user_profile_svc = UserProfileService(metadata)
        response: dict = user_profile_svc.disable_mfa(params)
        return ParseDict(response, user_pb2.UserInfo())

    def confirm_mfa(self, request, context):
        params, metadata = self.parse_request(request, context)
        user_profile_svc = UserProfileService(metadata)
        response: dict = user_profile_svc.confirm_mfa(params)
        return ParseDict(response, user_pb2.UserInfo())

    def get(self, request, context):
        params, metadata = self.parse_request(request, context)
        user_profile_svc = UserProfileService(metadata)
        response: dict = user_profile_svc.get(params)
        return ParseDict(response, user_pb2.UserInfo())

    def get_workspaces(self, request, context):
        params, metadata = self.parse_request(request, context)
        user_profile_svc = UserProfileService(metadata)
        response: dict = user_profile_svc.get_workspaces(params)
        return self.dict_to_message(response)

    def get_workspace_groups(self, request, context):
        params, metadata = self.parse_request(request, context)
        user_profile_svc = UserProfileService(metadata)
        response: dict = user_profile_svc.get_workspace_groups(params)
        return self.dict_to_message(response)



from spaceone.core.pygrpc import BaseAPI
from spaceone.api.identity.v2 import provider_pb2, provider_pb2_grpc
from spaceone.identity.service.provider_service import ProviderService


class Provider(BaseAPI, provider_pb2_grpc.ProviderServicer):
    pb2 = provider_pb2
    pb2_grpc = provider_pb2_grpc

    def create(self, request, context):
        params, metadata = self.parse_request(request, context)
        provider_svc = ProviderService(metadata)
        response: dict = provider_svc.create(params)
        return self.dict_to_message(response)

    def update(self, request, context):
        params, metadata = self.parse_request(request, context)
        provider_svc = ProviderService(metadata)
        response: dict = provider_svc.update(params)
        return self.dict_to_message(response)

    def update_plugin(self, request, context):
        params, metadata = self.parse_request(request, context)
        provider_svc = ProviderService(metadata)
        response: dict = provider_svc.update_plugin(params)
        return self.dict_to_message(response)

    def delete(self, request, context):
        params, metadata = self.parse_request(request, context)
        service_account_svc = ProviderService(metadata)
        service_account_svc.delete(params)
        return self.empty()

    def get(self, request, context):
        params, metadata = self.parse_request(request, context)
        provider_svc = ProviderService(metadata)
        response: dict = provider_svc.get(params)
        return self.dict_to_message(response)

    def list(self, request, context):
        params, metadata = self.parse_request(request, context)
        provider_svc = ProviderService(metadata)
        response: dict = provider_svc.list(params)
        return self.dict_to_message(response)

    def stat(self, request, context):
        params, metadata = self.parse_request(request, context)
        provider_svc = ProviderService(metadata)
        response: dict = provider_svc.stat(params)
        return self.dict_to_message(response)



from spaceone.core.pygrpc import BaseAPI
from spaceone.api.identity.plugin import (
    account_collector_pb2,
    account_collector_pb2_grpc,
)
from spaceone.identity.plugin.account_collector.service.account_collector_service import (
    AccountCollectorService,
)


class AccountCollector(BaseAPI, account_collector_pb2_grpc.AccountCollectorServicer):
    pb2 = account_collector_pb2
    pb2_grpc = account_collector_pb2_grpc

    def init(self, request, context):
        params, metadata = self.parse_request(request, context)
        account_collector_svc = AccountCollectorService(metadata)
        response: dict = account_collector_svc.init(params)
        return self.dict_to_message(response)

    def sync(self, request, context):
        params, metadata = self.parse_request(request, context)
        account_collector_svc = AccountCollectorService(metadata)
        response: dict = account_collector_svc.sync(params)
        return self.dict_to_message(response)



import logging
from datetime import datetime
from typing import Dict, List, Union

from spaceone.core.error import *
from spaceone.core.service import *
from spaceone.core.service.utils import *

from spaceone.identity.manager import SecretManager
from spaceone.identity.manager.domain_manager import DomainManager
from spaceone.identity.manager.project_group_manager import ProjectGroupManager
from spaceone.identity.manager.project_manager import ProjectManager
from spaceone.identity.manager.resource_manager import ResourceManager
from spaceone.identity.manager.role_binding_manager import RoleBindingManager
from spaceone.identity.manager.service_account_manager import ServiceAccountManager
from spaceone.identity.manager.trusted_account_manager import TrustedAccountManager
from spaceone.identity.manager.workspace_group_manager import WorkspaceGroupManager
from spaceone.identity.manager.workspace_manager import WorkspaceManager
from spaceone.identity.model import Workspace
from spaceone.identity.model.workspace.request import *
from spaceone.identity.model.workspace.response import *
from spaceone.identity.service.role_binding_service import RoleBindingService

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class WorkspaceService(BaseService):
    resource = "Workspace"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.domain_mgr = DomainManager()
        self.resource_mgr = ResourceManager()
        self.workspace_mgr = WorkspaceManager()
        self.service_account_mgr = ServiceAccountManager()
        self.rb_mgr = RoleBindingManager()
        self.workspace_group_mgr = WorkspaceGroupManager()

    @transaction(permission="identity:Workspace.write", role_types=["DOMAIN_ADMIN"])
    @convert_model
    def create(self, params: WorkspaceCreateRequest) -> Union[WorkspaceResponse, dict]:
        """Create workspace
        Args:
            params (WorkspaceCreateRequest): {
                'name': 'str',          # required
                'tags': 'dict',
                'domain_id': 'str',     # injected from auth (required)
            }
        Returns:
            WorkspaceResponse:
        """

        params_data = params.dict()
        params_data["created_by"] = self.transaction.get_meta("authorization.user_id")

        workspace_vo = self.workspace_mgr.create_workspace(params_data)

        return WorkspaceResponse(**workspace_vo.to_dict())

    @transaction(permission="identity:Workspace.write", role_types=["DOMAIN_ADMIN"])
    @convert_model
    def update(self, params: WorkspaceUpdateRequest) -> Union[WorkspaceResponse, dict]:
        """Update workspace
        Args:
            params (WorkspaceUpdateRequest): {
                'workspace_id': 'str',  # required
                'name': 'str',
                'tags': 'dict'
                'domain_id': 'str'      # injected from auth (required)
            }
        Returns:
            WorkspaceResponse:
        """
        workspace_vo = self.workspace_mgr.get_workspace(
            params.workspace_id, params.domain_id
        )

        # Check is managed resource
        self.resource_mgr.check_is_managed_resource_by_trusted_account(workspace_vo)

        workspace_vo = self.workspace_mgr.update_workspace_by_vo(
            params.dict(exclude_unset=True), workspace_vo
        )
        return WorkspaceResponse(**workspace_vo.to_dict())

    @transaction(permission="identity:Workspace.write", role_types=["DOMAIN_ADMIN"])
    @convert_model
    def change_workspace_group(
        self, params: WorkspaceChangeWorkspaceGroupRequest
    ) -> Union[WorkspaceResponse, dict]:
        """Change workspace group
        Args:
            params (WorkspaceChangeWorkspaceGroupRequest): {
                'workspace_id': 'str',                   # required
                'workspace_group_id': 'str',
                'domain_id': 'str'                       # injected from auth (required)
            }
        Returns:
            WorkspaceResponse:
        """
        workspace_id = params.workspace_id
        new_workspace_group_id = params.workspace_group_id
        domain_id = params.domain_id

        workspace_vo = self.workspace_mgr.get_workspace(
            workspace_id=params.workspace_id, domain_id=domain_id
        )

        old_workspace_group_id = workspace_vo.workspace_group_id
        is_updatable = True
        workspace_group_vo = None
        if new_workspace_group_id:
            workspace_group_vo = self.workspace_group_mgr.get_workspace_group(
                domain_id, new_workspace_group_id
            )
            is_updatable = self._add_workspace_to_group(
                workspace_id, new_workspace_group_id, domain_id
            )
        elif old_workspace_group_id:
            workspace_group_vo = self.workspace_group_mgr.get_workspace_group(
                domain_id, old_workspace_group_id
            )
            self._remove_workspace_from_group_with_workspace_vo(
                workspace_vo, old_workspace_group_id, domain_id
            )

        if is_updatable:
            workspace_vo = self.workspace_mgr.update_workspace_by_vo(
                params.dict(exclude_unset=False), workspace_vo
            )

            if new_workspace_group_id:
                workspace_vos = self.workspace_mgr.filter_workspaces(
                    workspace_group_id=new_workspace_group_id, domain_id=domain_id
                )

                self.workspace_group_mgr.update_workspace_group_by_vo(
                    {"workspace_count": len(workspace_vos)}, workspace_group_vo
                )
            if old_workspace_group_id:
                workspace_vos = self.workspace_mgr.filter_workspaces(
                    workspace_group_id=old_workspace_group_id, domain_id=domain_id
                )

                workspace_group_vo = self.workspace_group_mgr.get_workspace_group(
                    domain_id, old_workspace_group_id
                )

                self.workspace_group_mgr.update_workspace_group_by_vo(
                    {"workspace_count": len(workspace_vos)}, workspace_group_vo
                )

        return WorkspaceResponse(**workspace_vo.to_dict())

    @transaction(permission="identity:Workspace.write", role_types=["DOMAIN_ADMIN"])
    @convert_model
    def delete(self, params: WorkspaceDeleteRequest) -> None:
        """Delete workspace
        Args:
            params (WorkspaceDeleteRequest): {
                'force': 'bool',
                'workspace_id': 'str',  # required
                'domain_id': 'str'      # injected from auth (required)
            }
        Returns:
            None
        """

        domain_id = params.domain_id
        workspace_id = params.workspace_id

        workspace_vo = self.workspace_mgr.get_workspace(workspace_id, domain_id)

        # Check is managed resource
        self.resource_mgr.check_is_managed_resource_by_trusted_account(workspace_vo)

        service_account_vos = self.service_account_mgr.filter_service_accounts(
            domain_id=domain_id, workspace_id=workspace_id
        )

        if params.force:
            self._delete_related_resources_in_workspace(workspace_vo)
        elif len(service_account_vos) > 0:
            raise ERROR_UNKNOWN(
                message=f"Please delete service accounts in workspace ({workspace_id})"
            )
        else:
            self._delete_related_resources_in_workspace(workspace_vo)

        self.workspace_mgr.delete_workspace_by_vo(workspace_vo)

    @transaction(permission="identity:Workspace.write", role_types=["DOMAIN_ADMIN"])
    @convert_model
    def enable(self, params: WorkspaceEnableRequest) -> Union[WorkspaceResponse, dict]:
        """Enable workspace
        Args:
            params (WorkspaceEnableRequest): {
                'workspace_id': 'str',  # required
                'domain_id': 'str'      # injected from auth (required)
            }
        Returns:
            WorkspaceResponse:
        """
        workspace_vo = self.workspace_mgr.get_workspace(
            params.workspace_id, params.domain_id
        )

        # Check is managed resource
        self.resource_mgr.check_is_managed_resource_by_trusted_account(workspace_vo)

        workspace_vo = self.workspace_mgr.enable_workspace(workspace_vo)
        return WorkspaceResponse(**workspace_vo.to_dict())

    @transaction(permission="identity:Workspace.write", role_types=["DOMAIN_ADMIN"])
    @convert_model
    def disable(
        self, params: WorkspaceDisableRequest
    ) -> Union[WorkspaceResponse, dict]:
        """Disable workspace
        Args:
            params (WorkspaceDisableRequest): {
                'workspace_id': 'str',  # required
                'domain_id': 'str'      # injected from auth (required)
            }
        Returns:
            WorkspaceResponse:
        """

        workspace_vo = self.workspace_mgr.get_workspace(
            params.workspace_id, params.domain_id
        )

        # Check is managed resource
        self.resource_mgr.check_is_managed_resource_by_trusted_account(workspace_vo)

        workspace_vo = self.workspace_mgr.disable_workspace(workspace_vo)
        return WorkspaceResponse(**workspace_vo.to_dict())

    @transaction(permission="identity:Workspace.read", role_types=["DOMAIN_ADMIN"])
    @convert_model
    def get(self, params: WorkspaceGetRequest) -> Union[WorkspaceResponse, dict]:
        """Get workspace
        Args:
            params (WorkspaceGetRequest): {
                'workspace_id': 'str',  # required
                'domain_id': 'str'      # injected from auth (required)
            }
        Returns:
            WorkspaceResponse:
        """

        workspace_vo = self.workspace_mgr.get_workspace(
            params.workspace_id, params.domain_id
        )
        return WorkspaceResponse(**workspace_vo.to_dict())

    @transaction(role_types=["INTERNAL"])
    @convert_model
    def check(self, params: WorkspaceCheckRequest) -> None:
        """Check workspace
        Args:
            params (WorkspaceCheckRequest): {
                'workspace_id': 'str',  # required
                'domain_id': 'str'      # required
            }
        Returns:
            None:
        """

        self.workspace_mgr.get_workspace(params.workspace_id, params.domain_id)

    @transaction(permission="identity:Workspace.read", role_types=["DOMAIN_ADMIN"])
    @append_query_filter(
        [
            "workspace_id",
            "name",
            "state",
            "created_by",
            "is_managed",
            "is_dormant",
            "domain_id",
        ]
    )
    @append_keyword_filter(["workspace_id", "name"])
    @convert_model
    def list(
        self, params: WorkspaceSearchQueryRequest
    ) -> Union[WorkspacesResponse, dict]:
        """List workspaces
        Args:
            params (WorkspaceSearchQueryRequest): {
                'query': 'dict (spaceone.api.core.v1.Query)',
                'workspace_id': 'str',
                'name': 'str',
                'state': 'str',
                'created_by': 'str',
                'is_managed': 'bool',
                'is_dormant': 'bool',
                'workspace_group_id': 'str',
                'domain_id': 'str',         # injected from auth (required)
            }
        Returns:
            WorkspacesResponse:
        """

        query = params.query or {}
        workspace_group_id = params.workspace_group_id

        if not workspace_group_id:
            workspace_vos, total_count = self.workspace_mgr.list_workspaces(query)

            workspaces_info = [workspace_vo.to_dict() for workspace_vo in workspace_vos]
        else:
            workspaces_info, total_count = (
                self.workspace_mgr.list_workspace_group_workspaces(
                    params.workspace_group_id,
                    params.domain_id,
                )
            )

        return WorkspacesResponse(results=workspaces_info, total_count=total_count)

    @transaction(permission="identity:Workspace.read", role_types=["DOMAIN_ADMIN"])
    @append_query_filter(["domain_id"])
    @append_keyword_filter(["workspace_id", "name"])
    @convert_model
    def stat(self, params: WorkspaceStatQueryRequest) -> dict:
        """Stat workspaces
        Args:
            params (dict): {
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)', # required
                'domain_id': 'str',         # injected from auth (required)
            }
        Returns:
            dict: {
                'results': 'list',
                'total_count': 'int'
            }
        """

        query = params.query or {}
        return self.workspace_mgr.stat_workspaces(query)

    @staticmethod
    def _delete_related_resources_in_workspace(workspace_vo: Workspace):
        project_group_mgr = ProjectGroupManager()
        project_mgr = ProjectManager()
        trusted_account_mgr = TrustedAccountManager()
        service_account_mgr = ServiceAccountManager()
        secret_mgr = SecretManager()

        project_group_vos = project_group_mgr.filter_project_groups(
            domain_id=workspace_vo.domain_id, workspace_id=workspace_vo.workspace_id
        )

        project_vos = project_mgr.filter_projects(
            domain_id=workspace_vo.domain_id, workspace_id=workspace_vo.workspace_id
        )

        trusted_account_vos = trusted_account_mgr.filter_trusted_accounts(
            domain_id=workspace_vo.domain_id, workspace_id=workspace_vo.workspace_id
        )

        service_account_vos = service_account_mgr.filter_service_accounts(
            domain_id=workspace_vo.domain_id, workspace_id=workspace_vo.workspace_id
        )

        _LOGGER.debug(
            f"[_delete_related_resources_in_workspace] Start delete related resources in workspace: {workspace_vo.domain_id} {workspace_vo.name}( {workspace_vo.workspace_id} )"
        )

        if project_group_vos:
            _LOGGER.debug(
                f"[_delete_related_resources_in_workspace] Delete project groups count {workspace_vo.workspace_id} : {project_group_vos.count()}"
            )
            project_group_vos.delete()

        if project_vos:
            _LOGGER.debug(
                f"[_delete_related_resources_in_workspace] Delete projects count  {workspace_vo.workspace_id} : {project_vos.count()}"
            )
            project_vos.delete()

        for service_account_vo in service_account_vos:
            secret_mgr.delete_related_secrets(service_account_vo.service_account_id)
            _LOGGER.debug(
                f"[_delete_related_resources_in_workspace] Delete service account: {service_account_vo.name} ({service_account_vo.service_account_id})"
            )
            service_account_mgr.delete_service_account_by_vo(service_account_vo)

        for trusted_account_vo in trusted_account_vos:
            secret_mgr.delete_related_trusted_secrets(
                trusted_account_vo.trusted_account_id
            )
            _LOGGER.debug(
                f"[_delete_related_resources_in_workspace] Delete trusted account: {trusted_account_vo.name} ({trusted_account_vo.trusted_account_id})"
            )
            trusted_account_mgr.delete_trusted_account_by_vo(trusted_account_vo)

    def _add_workspace_to_group(
        self, workspace_id: str, workspace_group_id: str, domain_id: str
    ) -> bool:
        workspace_vo = self.workspace_mgr.get_workspace(
            workspace_id=workspace_id, domain_id=domain_id
        )
        old_workspace_group_id = workspace_vo.workspace_group_id
        is_updatable = True

        workspace_group_vo = self.workspace_group_mgr.get_workspace_group(
            domain_id, workspace_group_id
        )

        if old_workspace_group_id:
            if old_workspace_group_id != workspace_group_id:
                self._delete_role_bindings(
                    workspace_id, domain_id, old_workspace_group_id
                )

                self._create_role_bindings(
                    workspace_group_vo.users,
                    workspace_id,
                    workspace_group_id,
                    domain_id,
                )

                workspaces_info, total_count = (
                    self.workspace_mgr.list_workspace_group_workspaces(
                        old_workspace_group_id,
                        domain_id,
                    )
                )
                for workspace_info in workspaces_info:
                    workspace_vo = self.workspace_mgr.get_workspace(
                        workspace_info["workspace_id"], domain_id
                    )
                    user_rb_ids = self.rb_mgr.stat_role_bindings(
                        query={
                            "distinct": "user_id",
                            "filter": [
                                {
                                    "k": "workspace_id",
                                    "v": workspace_info["workspace_id"],
                                    "o": "eq",
                                },
                                {"k": "domain_id", "v": domain_id, "o": "eq"},
                            ],
                        }
                    ).get("results", [])
                    user_rb_total_count = len(user_rb_ids)

                    workspace_vo = self.workspace_mgr.update_workspace_by_vo(
                        {"user_count": user_rb_total_count}, workspace_vo
                    )
                    workspace_info.update({"user_count": workspace_vo.user_count})
            else:
                is_updatable = False
        else:
            workspace_group_dict = workspace_group_vo.to_dict()
            users = workspace_group_dict.get("users", []) or []
            user_ids = [user.get("user_id") for user in users]
            self._delete_role_bindings(workspace_id, domain_id, user_ids=user_ids)
            self._create_role_bindings(
                workspace_group_vo.users,
                workspace_id,
                workspace_group_id,
                domain_id,
            )

        if is_updatable:
            workspace_vo.changed_at = datetime.utcnow()
            self.workspace_mgr.update_workspace_by_vo(
                {"changed_at": workspace_vo.changed_at}, workspace_vo
            )

        return is_updatable

    def _remove_workspace_from_group_with_workspace_vo(
        self, workspace_vo: Workspace, old_workspace_group_id: str, domain_id: str
    ) -> None:
        workspace_id = workspace_vo.workspace_id

        workspace_vo.changed_at = datetime.utcnow()
        workspace_vo.workspace_group_id = None

        user_rb_ids = self.rb_mgr.stat_role_bindings(
            query={
                "distinct": "user_id",
                "filter": [
                    {"k": "workspace_id", "v": workspace_id, "o": "eq"},
                    {"k": "domain_id", "v": domain_id, "o": "eq"},
                ],
            }
        ).get("results", [])
        self._delete_role_bindings(
            workspace_id, domain_id, old_workspace_group_id, user_rb_ids
        )
        user_rb_total_count = len(user_rb_ids)

        self.workspace_mgr.update_workspace_by_vo(
            {
                "user_count": user_rb_total_count,
                "changed_at": workspace_vo.changed_at,
                "workspace_group_id": None,
            },
            workspace_vo,
        )

    def _delete_role_bindings(
        self,
        workspace_id: str,
        domain_id: str,
        existing_workspace_group_id: str = None,
        user_ids: List[str] = None,
    ):
        rb_vos = self.rb_mgr.filter_role_bindings(
            workspace_id=workspace_id,
            domain_id=domain_id,
            workspace_group_id=existing_workspace_group_id,
            user_id=user_ids,
        )
        for rb_vo in rb_vos:
            self.rb_mgr.delete_role_binding_by_vo(rb_vo)

    @staticmethod
    def _create_role_bindings(
        workspace_group_users: List[Dict[str, str]],
        workspace_id: str,
        workspace_group_id: str,
        domain_id: str,
    ):
        rb_svc = RoleBindingService()
        for user_info in workspace_group_users or []:
            rb_svc.create_role_binding(
                {
                    "user_id": user_info["user_id"],
                    "role_id": user_info["role_id"],
                    "resource_group": "WORKSPACE",
                    "domain_id": domain_id,
                    "workspace_group_id": workspace_group_id,
                    "workspace_id": workspace_id,
                }
            )



from mongoengine import *
from spaceone.core.model.mongo_model import MongoModel


class Schema(MongoModel):
    schema_id = StringField(max_length=40, unique_with="domain_id")
    name = StringField(max_length=40)
    version = StringField(max_length=40, default=None, null=True)
    schema_type = StringField(
        max_length=20,
        choices=("SERVICE_ACCOUNT", "TRUSTED_ACCOUNT", "SECRET", "TRUSTING_SECRET"),
    )
    schema = DictField(default=None)
    provider = StringField(max_length=40)
    related_schemas = ListField(StringField(max_length=40), default=None)
    options = DictField(default=None)
    tags = DictField(default=None)
    is_managed = BooleanField(default=False)
    domain_id = StringField(max_length=40)
    created_at = DateTimeField(auto_now_add=True)
    updated_at = DateTimeField(auto_now=True)

    meta = {
        "updatable_fields": [
            "name",
            "version",
            "schema",
            "related_schemas",
            "options",
            "tags",
            "updated_at",
        ],
        "minimal_fields": [
            "schema_id",
            "name",
            "schema_type",
            "provider",
            "is_managed",
            "version",
        ],
        "change_query_keys": {
            "related_schema_id": "related_schemas",
        },
        "ordering": ["name"],
        "indexes": [
            "schema_id",
            "schema_type",
            "provider",
            "related_schemas",
            "is_managed",
            "domain_id",
        ],
    }



import logging
import os
from typing import List

from spaceone.core import utils
from spaceone.core.manager import BaseManager

_LOGGER = logging.getLogger(__name__)
CURRENT_DIR = os.path.dirname(__file__)
_PROVIDER_DIR = os.path.join(CURRENT_DIR, "../managed_resource/provider/")
_SCHEMA_DIR = os.path.join(CURRENT_DIR, "../managed_resource/schema/")
_ROLE_DIR = os.path.join(CURRENT_DIR, "../managed_resource/role/")


class ManagedResourceManager(BaseManager):

    def get_managed_providers(self) -> dict:
        provider_map = {}
        for provider in self._load_managed_resources(_PROVIDER_DIR):
            provider_map[provider["provider"]] = provider

        return provider_map

    def get_managed_schemas(self) -> dict:
        schema_map = {}
        for schema in self._load_managed_resources(_SCHEMA_DIR):
            schema_map[schema["schema_id"]] = schema

        return schema_map

    def get_managed_roles(self) -> dict:
        role_map = {}
        for role in self._load_managed_resources(_ROLE_DIR):
            role_map[role["role_id"]] = role

        return role_map

    @staticmethod
    def _load_managed_resources(dir_path: str) -> List[dict]:
        managed_resources = []
        for filename in os.listdir(dir_path):
            if filename.endswith(".yaml"):
                file_path = os.path.join(dir_path, filename)
                managed_resource_info = utils.load_yaml_from_file(file_path)
                managed_resources.append(managed_resource_info)
        return managed_resources
